/* Copyright 2013 Andrey Dubovik <andrei.dubovik@gmail.com>
 *
 * This file is part of Minido.
 *
 * Minido is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Minido is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Minido. If not, see <http://www.gnu.org/licenses/>.
 */

#include "database.hpp"
#include "unicode.hpp"

using namespace std;

Icon::Icon(): icon(nullptr) {}

Icon::Icon(const string& name) {
  GtkIconTheme* icon_theme = gtk_icon_theme_get_default();
  icon = gtk_icon_theme_load_icon(icon_theme, name.c_str(), 48,
				  (GtkIconLookupFlags)0, NULL);
}

Icon::~Icon() {
  if(icon) g_object_unref(icon);
}

Icon::Icon(Icon&& rhs) {
  icon = rhs.icon;
  rhs.icon = nullptr;
}

Icon& Icon::operator=(Icon&& rhs) {
  if(icon) g_object_unref(icon);
  icon = rhs.icon;
  rhs.icon = nullptr;
}

Icon::operator GdkPixbuf*() const {
  return icon;
}

Item::Item(DataBase& db): db(db), tick(0) {
  db.items.push_back(this);
  id = db.items.size();
  ++db.size;
}

Item::~Item() {
  db.items[id-1] = nullptr;
  --db.size;
}

void Item::setName(const u32string& name) {
  Item::name = name;
  lcname = downcase(name);
}

void Item::setName(const u32string&& name) {
  lcname = downcase(name);
  Item::name = move(name);
}

const u32string& Item::getName() const {
  return name;
}

LScore::LScore(): alpha(0), beta(0) {}

bool LScore::operator<(const LScore& rhs) const {
  if(alpha < rhs.alpha) return true;
  if(rhs.alpha < alpha) return false;
  if(beta < rhs.beta) return true;
  return false;
}

GScore::GScore(): alpha(0), beta(0), length(0) {}

GScore::GScore(const LScore& lscore):
  alpha(lscore.alpha),
  beta(lscore.beta),
  length(0) {}

bool GScore::operator<(const GScore& rhs) const {
  if(alpha < rhs.alpha) return true;
  if(rhs.alpha < alpha) return false;
  if(similar < rhs.similar) return true;
  if(rhs.similar < similar) return false;
  if(tick < rhs.tick) return true;
  if(rhs.tick < tick) return false;
  if(beta < rhs.beta) return true;
  if(rhs.beta < beta) return false;
  if(length < rhs.length) return true;
  return false;
}

LScore computeScore(const u32string& name,
		   const u32string& pattern,
		   int nc, int pc, int alpha,
		   int sublen, int beta)
{
  LScore bscore;

  if(pc == pattern.size()) {
    alpha += sublen * sublen;
    bscore.alpha = alpha;
    bscore.beta = beta;
    return bscore;
  }

  bool match = false;

  for(int ni = nc; ni < name.size(); ++ni) {
    if(name[ni] == pattern[pc]) {
      match = true;
      if(!pc) beta = -ni;
      LScore score;
      if(ni == nc) {
	score = computeScore(name, pattern, ni+1, pc+1,
				   alpha, sublen + 1, beta);
	if(bscore < score) bscore = score;
      }
      else {
	score = computeScore(name, pattern, ni+1, pc+1,
				   alpha + sublen * sublen, 1, beta);
      }
      if(bscore < score) bscore = score;
    }
  }
  if(!match) return LScore();
  return bscore;
}

LScore computeMask(const u32string& name,
		   const u32string& pattern,
		   MatchMask& mask,
		   int nc, int pc, int alpha,
		   int sublen, int beta)
{
  if(pc == 0) {
    mask.resize(name.size(), 0);
  }

  auto bmask = mask;
  LScore bscore;

  if(pc == pattern.size()) {
    alpha += sublen * sublen;
    bscore.alpha = alpha;
    bscore.beta = beta;
    return bscore;
  }

  bool match = false;

  for(int ni = nc; ni < name.size(); ++ni) {
    if(name[ni] == pattern[pc]) {
      match = true;
      if(!pc) beta = -ni;
      auto fmask = mask;
      fmask[ni] = 1;

      LScore score;
      if(ni == nc) {
	score = computeMask(name, pattern, fmask, ni+1, pc+1,
				   alpha, sublen + 1, beta);
      }
      else {
	score = computeMask(name, pattern, fmask, ni+1, pc+1,
				   alpha + sublen * sublen, 1, beta);
      }
      if(bscore < score) {
	bscore = score;
	bmask = move(fmask);
      }
    }
  }
  if(!match) return LScore();
  mask = move(bmask);
  return bscore;
}

SearchResult::SearchResult(): item(nullptr) {}

SearchResult::operator bool() const {
  if(item) return true;
  return false;
}

DataBase::DataBase(): size(0), tick(0) {}

SearchResult DataBase::search(const u32string& pattern) {
  SearchResult result;
  if(!size) return result;

  auto lcpatt = downcase(pattern);
  
  GScore bscore;
  for(auto& item: items) {
    if(item) {
      GScore score = computeScore(item->lcname, lcpatt);
      score.length = -item->lcname.size();
      score.tick = item->tick;
      if(item->lcpatt.size()) {
        score.similar = isSimilar(lcpatt, item->lcpatt);
      }
      if(bscore < score && score.alpha) {
        bscore = score;
        result.item = item;
      }
    }
  }

  if(result.item) {
    computeMask(result.item->lcname, lcpatt, result.mask);
  }
  return result;
}

void Item::invoke() {

}

void Item::invoke(const u32string& patt) {
  lcpatt = downcase(patt);
  tick = ++db.tick;
  invoke();
}

LScore isSimilar(const u32string& str1, const u32string& str2) {
  if(str1.size() < str2.size()) return isSimilar(str2, str1);
  if(!str2.size()) return LScore();

  LScore score;

  int cur = str1.find(str2[0]);
  if(cur == u32string::npos) return score;
  score.beta = -cur;

  for(auto c2 = ++str2.cbegin(); c2 != str2.cend(); ++c2) {
    cur = str1.find(*c2, cur);
    if(cur == u32string::npos) return score;
    ++cur;
  }

  score.alpha = 1;
  return score;
}

