/* Copyright 2013 Andrey Dubovik <andrei.dubovik@gmail.com>
 *
 * This file is part of Minido.
 *
 * Minido is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Minido is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Minido. If not, see <http://www.gnu.org/licenses/>.
 */

#include <gtk/gtk.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <string>
#include <dlfcn.h>
#include <cmath>

#include "unicode.hpp"
#include "utilities.hpp"
#include "database.hpp"

using namespace std;

gchar* match_style;
gchar* info_style;
const char* info_message = "Type To Search";

static int focus = 0;
static int active_grab = 0;
static GThread* hotkey_thread;
static Display* xdisplay;

struct Minido {
  GtkStatusIcon* systray_icon;
  GtkWidget* search_window;
  GtkWidget* systray_menu;
  GtkWidget* about_dialog;
  GtkWidget* invisible;
  GtkWidget* search_input;
  GtkWidget* icon;
  GdkDevice* pointer;
  GdkDevice* keyboard;
};

static Minido minido;
static u32string search_string;
static SearchResult search_result;
static DataBase db;

static void* libapps;

gboolean grab_keyboard(gpointer data) {
  // Grap keyboard
  GdkWindow* gdk_invisible = gtk_widget_get_window(minido.invisible);
  gdk_device_grab(minido.keyboard, gdk_invisible, GDK_OWNERSHIP_NONE, TRUE, GDK_ALL_EVENTS_MASK, NULL, GDK_CURRENT_TIME);
  active_grab = 1;

  return FALSE; // remove the function from the main loop
}

gboolean show(gpointer data) {
  // Clean after last search
  search_string.clear();
  gtk_label_set_markup(GTK_LABEL(minido.search_input), decorate(info_message, info_style).c_str());
  gtk_image_set_from_stock(GTK_IMAGE(minido.icon), GTK_STOCK_FIND, GTK_ICON_SIZE_DIALOG);

  // Grab pointer
  GdkWindow* gdk_invisible = gtk_widget_get_window(minido.invisible);
  gdk_device_grab(minido.pointer, gdk_invisible, GDK_OWNERSHIP_NONE, TRUE, GDK_ALL_EVENTS_MASK, NULL, GDK_CURRENT_TIME);

  // Show Minido
  gtk_grab_add(minido.search_window);
  gtk_widget_show(minido.search_window);

  return FALSE; // remove the function from the main loop
}

gpointer hotkey_loop(gpointer data) {
  // Passively grab keyboard
  KeyCode keycode = XKeysymToKeycode(xdisplay, XK_space);
  Window root = DefaultRootWindow(xdisplay);
  Atom quit_atom = XInternAtom(xdisplay, "quit", False);
  XGrabKey(xdisplay, keycode, ControlMask, root, TRUE, GrabModeAsync, GrabModeAsync);
  XSelectInput(xdisplay, root, KeyPressMask | KeyReleaseMask);

  // Wait for KeyPress or quit signal
  XEvent event;
  while(1) {
    XNextEvent(xdisplay, &event);
    if(event.type == ClientMessage && event.xclient.message_type == quit_atom) {
      XUngrabKey(xdisplay, keycode, ControlMask, root);
      return NULL;
    }
    if(event.type == KeyPress && event.xkey.keycode == keycode) {
      if(!gtk_widget_get_visible(minido.search_window)) {
	g_idle_add(&show, NULL);

	// Wait for KeyRelease before grabbing keyboard
	while(1) {
	  XNextEvent(xdisplay, &event);
	  if(event.type == KeyRelease && event.xkey.keycode == keycode) {
	    // Check for autorepeat
	    if(XQLength(xdisplay)) {
	      XNextEvent(xdisplay, &event);
	      if(event.type == KeyPress && event.xkey.keycode == keycode) {
		continue;
	      }
	    }
	    g_idle_add(&grab_keyboard, NULL);
	    break;
	  }
	}
      }
    }
  }
}


void hide() {
  // Release keyboard and pointer grab
  gdk_device_ungrab(minido.pointer, GDK_CURRENT_TIME);
  gdk_device_ungrab(minido.keyboard, GDK_CURRENT_TIME);
  active_grab = 0;

  // Hide Minido
  gtk_widget_hide(minido.search_window);
  gtk_grab_remove(minido.search_window);
}


void activate(GtkApplication* application, gpointer data) {
  hotkey_thread = g_thread_new("hotkey_loop", &hotkey_loop, NULL);
}

gboolean button_press_event(GtkWidget* widget, GdkEventButton* event, gpointer data) {
  if(!focus && active_grab) {
    hide();
  }
  return TRUE;
}

gboolean key_press_event(GtkWidget* widget, GdkEventKey* event, gpointer data) {
  gunichar c32 = gdk_keyval_to_unicode(event->keyval);
  if(event->keyval == GDK_KEY_Return) {
    if(search_string.size() && search_result) {
      search_result.item->invoke(search_string);
      hide();
      return TRUE;
    }
    else return TRUE;
  }
  else if(event->keyval == GDK_KEY_Escape) {
    if(!search_string.size()) {
      hide();
      return TRUE;
    }
    else {
      search_string.clear();
    }
  }
  else if(event->keyval == GDK_KEY_BackSpace) {
    if(search_string.size()) {
      search_string.pop_back();
    }
    else return TRUE;
  }
  else if(c32 <= 32U || c32 == 127U) {
    return TRUE;
  }
  else {
    search_string.push_back((char32_t)c32);
  }

  GdkPixbuf* icon = nullptr;
  string fname;
  if(search_string.size()) {
    search_result = db.search(search_string);
    if(search_result) {
      fname = decorate(search_result.item->getName(), match_style, search_result.mask);
      icon = search_result.item->icon;
    }
    else {
      fname = decorate(search_string, info_style);
    }
  }
  else {
    fname = decorate(info_message, info_style);
  }

  gtk_label_set_markup(GTK_LABEL(minido.search_input), fname.c_str());

  // Display icon
  if(icon) {
    gtk_image_set_from_pixbuf(GTK_IMAGE(minido.icon), icon);
  }
  else {
    gtk_image_set_from_stock(GTK_IMAGE(minido.icon), GTK_STOCK_FIND, GTK_ICON_SIZE_DIALOG);
  }

  return TRUE;
}

gboolean enter_notify_event(GtkWidget* widget, GdkEventCrossing* event, gpointer data) {
  if(event->window == gtk_widget_get_window(minido.search_window)) {
    focus = 1;
  }
  return TRUE;
}

gboolean leave_notify_event(GtkWidget* widget, GdkEventCrossing* event, gpointer data) {
  if(event->window == gtk_widget_get_window(minido.search_window)) {
    focus = 0;
  }
  return TRUE;
}

void systray_activate(GtkStatusIcon* icon, gpointer data) {
  show(NULL);
  grab_keyboard(NULL);
}

void systray_popup(GtkStatusIcon* icon, guint button, guint activate_time, gpointer data) {
  gtk_menu_popup(GTK_MENU(minido.systray_menu), NULL, NULL, gtk_status_icon_position_menu, icon, button, activate_time);
}

void menu_quit(GtkMenuItem* menuitem, gpointer data) {
  // Ask the hotkey thread to quit
  Window root = DefaultRootWindow(xdisplay);
  Atom quit_atom = XInternAtom(xdisplay, "quit", False);

  XEvent event;
  event.xclient.type = ClientMessage;
  event.xclient.display = xdisplay;
  event.xclient.window = root;
  event.xclient.message_type = quit_atom;
  event.xclient.format = 8;
  XSendEvent(xdisplay, root, False, KeyPressMask, &event);
  XFlush(xdisplay);

  // Wait for the hotkey thread to finish
  g_thread_join(hotkey_thread);
  
  // Free memory
  g_object_unref(minido.systray_icon);
  g_object_unref(minido.systray_menu);
  gtk_widget_destroy(minido.invisible);
  gtk_widget_destroy(minido.about_dialog);
  gtk_widget_destroy(minido.search_window);
  g_free(match_style);
  g_free(info_style);

  // Disconnect from X
  XCloseDisplay(xdisplay);
  
  // Plugin Tests
  dlclose(libapps);
}

void menu_about(GtkMenuItem* menuitem, gpointer data) {
  gtk_dialog_run(GTK_DIALOG(minido.about_dialog));
  gtk_widget_hide(minido.about_dialog);
}

gchar* get_input_markup(const gchar* region_name) {
  // Initilize style context
  GtkStyleContext* context = gtk_style_context_new();
  GtkWidgetPath* path = gtk_widget_get_path(minido.search_input);
  gtk_style_context_set_path(context, path);
  gtk_style_context_add_region(context, region_name, GTK_REGION_ONLY);
  gchar* style;
  
  // Base font
  GtkStyleContext* b_context = gtk_widget_get_style_context(minido.search_input);
  const PangoFontDescription* b_font = gtk_style_context_get_font(b_context, GTK_STATE_FLAG_NORMAL);
  char* b_font_str = pango_font_description_to_string(b_font);

  // Font
  const PangoFontDescription* font = gtk_style_context_get_font(context, GTK_STATE_FLAG_NORMAL);
  char* font_str = pango_font_description_to_string(font);

  // Take care of kerning (do not add a font attribute when not necessary)
  if(g_strcmp0(b_font_str, font_str) == 0) {
    style = g_strconcat("", NULL);
  }
  else {
    style = g_strconcat("font='", font_str, "' ", NULL);
  }
  g_free(b_font_str);
  g_free(font_str);


  gchar* buf;

  // Foreground color
  GdkRGBA color;
  gtk_style_context_get_color(context, GTK_STATE_FLAG_NORMAL, &color);
  gchar* color_str = g_strdup_printf("#%02X%02X%02X", (int)round(color.red * 255), (int)round(color.green * 255), (int)round(color.blue * 255));
  buf = g_strconcat(style, "color='", color_str, "' ", NULL);
  g_free(style);
  g_free(color_str);
  style = buf;

  // Background color
  GdkRGBA bgcolor;
  gtk_style_context_get_background_color(context, GTK_STATE_FLAG_NORMAL, &bgcolor);
  if((int)round(bgcolor.alpha * 255) != 255) {
    gchar* bgcolor_str = g_strdup_printf("#%02X%02X%02X", (int)round(bgcolor.red * 255), (int)round(bgcolor.green * 255), (int)round(bgcolor.blue * 255));
    buf = g_strconcat(style, "bgcolor='", bgcolor_str, "' ", NULL);
    g_free(style);
    g_free(bgcolor_str);
    style = buf;
  }

  // Underline
  GValue underline = G_VALUE_INIT;
  gtk_style_context_get_property(context, "-pango-underline", GTK_STATE_FLAG_NORMAL, &underline);
  GEnumValue* underline_enum = g_enum_get_value(G_ENUM_CLASS(g_type_class_peek_static(PANGO_TYPE_UNDERLINE)),
						g_value_get_enum(&underline));
  buf = g_strconcat(style, "underline='", underline_enum->value_nick, "' ", NULL);
  g_free(style);
  style = buf;

  // Clean up & return
  g_object_unref(context);
  return style;
}


void startup(GtkApplication* application, gpointer data) {
  // Register custom CSS properties // Depreciated since 3.8, but no 'text-decoration' property...
  GParamSpec* pango_underline = g_param_spec_enum("underline", "underline", "Pango Underline",
						  PANGO_TYPE_UNDERLINE, PANGO_UNDERLINE_NONE, G_PARAM_READABLE);
  gtk_theming_engine_register_property("pango", NULL, pango_underline);

  // Load UI
  GtkBuilder* builder = gtk_builder_new();
  gtk_builder_add_from_file(builder, DATADIR "/minido.ui", NULL);
  minido.search_window = GTK_WIDGET(gtk_builder_get_object(builder, "window"));

  // Add CSS styles
  GtkCssProvider* provider = gtk_css_provider_new();
  GdkScreen* screen = gdk_screen_get_default();
  gtk_css_provider_load_from_path(provider, DATADIR "/minido.css", NULL);
  gtk_style_context_add_provider_for_screen(screen, GTK_STYLE_PROVIDER(provider), GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
  g_object_unref(provider);

  // Style search input as an entry
  minido.search_input = GTK_WIDGET(gtk_builder_get_object(builder, "input"));
  GtkStyleContext* context = gtk_widget_get_style_context(minido.search_input);
  gtk_style_context_add_class(context, "entry");

  // Prepare pango styles
  match_style = get_input_markup("selected");
  info_style = get_input_markup("inactive");

  // Set initial greeting
  gtk_label_set_markup(GTK_LABEL(minido.search_input), decorate(info_message, info_style).c_str());

  // Initialize devices
  GdkDisplay* display = gdk_display_get_default();
  GdkDeviceManager* manager = gdk_display_get_device_manager(display);
  minido.pointer = gdk_device_manager_get_client_pointer(manager);
  minido.keyboard = gdk_device_get_associated_device(minido.pointer);

  // Construct GtkInvisible (for device grabbing)
  minido.invisible = gtk_invisible_new();
  gtk_widget_show(minido.invisible);

  // Load icon
  minido.icon =  GTK_WIDGET(gtk_builder_get_object(builder, "icon"));

  // Connect signals
  g_signal_connect(minido.search_window, "enter-notify-event", G_CALLBACK(enter_notify_event), NULL);
  g_signal_connect(minido.search_window, "leave-notify-event", G_CALLBACK(leave_notify_event), NULL);
  g_signal_connect(minido.search_window, "key-press-event", G_CALLBACK(key_press_event), NULL);
  g_signal_connect(minido.search_window, "button-press-event", G_CALLBACK(button_press_event), NULL);

  // Create a systray icon
  minido.systray_icon = gtk_status_icon_new_from_stock(GTK_STOCK_FIND);
  gtk_status_icon_set_tooltip_text(minido.systray_icon, "Minido");
  g_signal_connect(minido.systray_icon, "activate", G_CALLBACK(systray_activate), NULL);
  g_signal_connect(minido.systray_icon, "popup-menu", G_CALLBACK(systray_popup), NULL);

  // Construct context menu
  minido.systray_menu = GTK_WIDGET(gtk_builder_get_object(builder, "menu"));
  g_object_ref(minido.systray_menu);
  GtkWidget* about_item = GTK_WIDGET(gtk_builder_get_object(builder, "about"));
  GtkWidget* quit_item = GTK_WIDGET(gtk_builder_get_object(builder, "quit"));
  gtk_menu_item_set_label(GTK_MENU_ITEM(about_item), "About Minido");
  gtk_menu_item_set_label(GTK_MENU_ITEM(quit_item), "Quit");
  
  g_signal_connect(about_item, "activate", G_CALLBACK(menu_about), NULL);
  g_signal_connect(quit_item, "activate", G_CALLBACK(menu_quit), NULL);

  // Load about dialog
  minido.about_dialog = GTK_WIDGET(gtk_builder_get_object(builder, "about_dialog"));

  // Register top windows with the application
  gtk_application_add_window(application, GTK_WINDOW(minido.search_window));
  gtk_application_add_window(application, GTK_WINDOW(minido.about_dialog));

  // Connect to X (for hotkeys)
  xdisplay = XOpenDisplay(NULL);

  // Free memory
  g_object_unref(builder);
  
  // Plugin Tests
  libapps = dlopen(PLUGINSDIR "/applications.so", RTLD_NOW);
  auto populate = (void(*)(DataBase& db))dlsym(libapps, "populate");
  populate(db);
}

int main(int argc, char** argv) {
  // Ask Xlib to be nice to my threads (xcb?)
  XInitThreads();

  // Construct GtkApplication
  GtkApplication* application = gtk_application_new("net.sourceforge.Minido", G_APPLICATION_FLAGS_NONE);
  //g_object_set(G_OBJECT(application), "register-session", TRUE, NULL); // Does not work on Xfce
  g_signal_connect(application, "startup", G_CALLBACK(startup), NULL);
  g_signal_connect(application, "activate", G_CALLBACK(activate), NULL);

  // Register the application
  g_application_register(G_APPLICATION(application), NULL, NULL);

  // Non-primary instance exits
  if(g_application_get_is_remote(G_APPLICATION(application))) {
  g_message("Minido is already running. Exiting..."); return 0; }

  // Run the application
  int status = g_application_run(G_APPLICATION(application), argc, argv);

  // Free memory
  g_object_unref(application);

  return status;
}
